//
//  ViewController.m
//  testReaderAsset
//
//  Created by Meicam on 2017/10/23.
//  Copyright © 2017年 Meicam. All rights reserved.
//

#import "ViewController.h"
#import <AVFoundation/AVFoundation.h>
#import <CoreImage/CoreImage.h>

#import <OpenGLES/EAGL.h>

@interface ViewController (){
    AVAssetReader *assetReader;
    AVAssetReaderTrackOutput *readerVideoTrackOutput, *readerAudioTrackOutput;
    AVAssetWriter *assetWriter;
    AVAssetWriterInput *assetVideoWriterInput, *assetAudioWriterInput;
}
@property (weak, nonatomic) IBOutlet UIImageView *imageView;

@end

@implementation ViewController{
    CIContext * context;
//    CIFilter *filter;
}

- (void)viewDidLoad {
    [super viewDidLoad];
    // Do any additional setup after loading the view, typically from a nib.
    EAGLContext *eaglctx = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
    context = [CIContext contextWithEAGLContext:eaglctx];
    
    NSLog(@"%@",NSHomeDirectory());
    NSString *path = [[NSBundle mainBundle] pathForResource:@"IMG_1781" ofType:@"mp4"];
    NSString *writerPath = [NSHomeDirectory() stringByAppendingPathComponent:@"Documents/abc.mp4"];
    NSFileManager *fileManager = [NSFileManager defaultManager];
    if ([fileManager fileExistsAtPath:writerPath]) {
        [fileManager removeItemAtPath:writerPath error:nil];
    }
    AVAsset *asset = [AVAsset assetWithURL:[NSURL fileURLWithPath:path]];
    
    AVAssetReaderOutput *readerVideoTrackOutput = nil;
    AVAssetReaderOutput *readerAudioTrackOutput = nil;

    /********************************************************************************************/

    // 通过 AVAssetReaderVideoCompositionOutput 配置
    assetReader = [self createAssetReader:asset];

    for( AVAssetReaderOutput *output in assetReader.outputs ) {
        if( [output.mediaType isEqualToString:AVMediaTypeAudio] ) {
            readerAudioTrackOutput = output;
        }
        else if( [output.mediaType isEqualToString:AVMediaTypeVideo] ) {
            readerVideoTrackOutput = output;
        }
    }
    
    //开始读
    [assetReader startReading];
    
    NSURL *writerUrl = [NSURL fileURLWithPath:writerPath];
    //创建一个写数据对象
    assetWriter = [[AVAssetWriter alloc] initWithURL:writerUrl fileType:AVFileTypeMPEG4 error:nil];
    //配置写数据，设置比特率，帧率等
    NSDictionary *compressionProperties = @{ AVVideoAverageBitRateKey : @(1.38*1024*1024),
                                             AVVideoExpectedSourceFrameRateKey: @(30),
                                             AVVideoProfileLevelKey : AVVideoProfileLevelH264HighAutoLevel };
    //配置编码器宽高等
    NSDictionary *compressionVideoSetting = @{
                              AVVideoCodecKey                   : AVVideoCodecTypeH264,
                              AVVideoWidthKey                   : @1280,
                              AVVideoHeightKey                  : @720,
                              AVVideoCompressionPropertiesKey   : compressionProperties
                              };

    AudioChannelLayout stereoChannelLayout = {
        .mChannelLayoutTag = kAudioChannelLayoutTag_Stereo,
        .mChannelBitmap = 0,
        .mNumberChannelDescriptions = 0
    };
    NSData *channelLayoutAsData = [NSData dataWithBytes:&stereoChannelLayout length:offsetof(AudioChannelLayout, mChannelDescriptions)];
    //写入音频配置
    NSDictionary *compressionAudioSetting = @{
                                               AVFormatIDKey         : [NSNumber numberWithUnsignedInt:kAudioFormatMPEG4AAC],
                                               AVEncoderBitRateKey   : [NSNumber numberWithInteger:64000],
                                               AVSampleRateKey       : [NSNumber numberWithInteger:44100],
                                               AVChannelLayoutKey    : channelLayoutAsData,
                                               AVNumberOfChannelsKey : [NSNumber numberWithUnsignedInteger:2]
                                               };
    //AVAssetWriterInput用来说明怎么写数据
    assetVideoWriterInput = [AVAssetWriterInput assetWriterInputWithMediaType:AVMediaTypeVideo outputSettings:compressionVideoSetting];
    assetAudioWriterInput = [AVAssetWriterInput assetWriterInputWithMediaType:AVMediaTypeAudio outputSettings:compressionAudioSetting];
    
    if ([assetWriter canAddInput:assetVideoWriterInput]) {
        [assetWriter addInput:assetVideoWriterInput];
    }
    if ([assetWriter canAddInput:assetAudioWriterInput]) {
        [assetWriter addInput:assetAudioWriterInput];
    }
    //开始写
    [assetWriter startWriting];
    [assetWriter startSessionAtSourceTime:kCMTimeZero];
    dispatch_group_t group = dispatch_group_create();
    dispatch_queue_t videoWriter = dispatch_queue_create("videoWriter", DISPATCH_QUEUE_CONCURRENT);
    dispatch_queue_t audioWriter = dispatch_queue_create("audioWriter", DISPATCH_QUEUE_CONCURRENT);
    __block BOOL isVideoComplete = NO;
    dispatch_group_enter(group);
    //要想写数据，就带有数据源，以下是将readerVideoTrackOutput读出来的数据加入到assetVideoWriterInput中再写入本地，音频和视频读取写入方式一样
    [assetVideoWriterInput requestMediaDataWhenReadyOnQueue:videoWriter usingBlock:^{
        while (!isVideoComplete && assetVideoWriterInput.isReadyForMoreMediaData) {
            //样本数据
            @autoreleasepool {
                //每次读取一个buffer
                CMSampleBufferRef buffer = [readerVideoTrackOutput copyNextSampleBuffer];
                if (buffer) {
                    //将读来的buffer加入到写对象中开始写
                    //此处也可以给assetVideoWriterInput加个适配器对象可以写入CVPixelBuffer
                    [assetVideoWriterInput appendSampleBuffer:buffer];
                    //将buffer生成图片
                    UIImage *image = [self imageFromSampleBuffer:buffer];
                    //给图片添加滤镜
                    UIImage *img = [self imageAddFilter:image];
                    dispatch_async(dispatch_get_main_queue(), ^{
                        self.imageView.image = img;
                    });
                    CFRelease(buffer);
                    buffer = NULL;
                } else {
                    isVideoComplete = YES;
                }
            }
            
        }
        if (isVideoComplete) {
            //关闭写入会话
            [assetVideoWriterInput markAsFinished];
            dispatch_group_leave(group);
        }
    }];
    __block BOOL isAudioComplete = NO;
    dispatch_group_enter(group);
    [assetAudioWriterInput requestMediaDataWhenReadyOnQueue:audioWriter usingBlock:^{
        while (!isAudioComplete && assetAudioWriterInput.isReadyForMoreMediaData) {
            //样本数据
            CMSampleBufferRef buffer = [readerAudioTrackOutput copyNextSampleBuffer];
            if (buffer) {
                [assetAudioWriterInput appendSampleBuffer:buffer];
                CFRelease(buffer);
                buffer = NULL;
            } else {
                isAudioComplete = YES;
            }

        }
        if (isAudioComplete) {
            //关闭写入会话
            [assetAudioWriterInput markAsFinished];
            dispatch_group_leave(group);
        }
    }];
    
    dispatch_group_notify(group, dispatch_get_main_queue(), ^{
        NSLog(@"OKKK");
        [assetWriter finishWritingWithCompletionHandler:^{
            AVAssetWriterStatus status = assetWriter.status;
            if (status == AVAssetWriterStatusCompleted) {
                [self exportMovieToPhotosAlbum:@[[NSURL fileURLWithPath:writerPath]]];
                
                NSLog(@"video finsished");
                [assetReader cancelReading];
                [assetWriter cancelWriting];
            } else {
                NSLog(@"video failure");
                NSLog(@"%@",assetWriter.error);
            }
            
        }];
    });

}

- (AVAssetReader*)createAssetReader:(AVAsset *)asset
{
    NSLog(@"creating reader from composition: %@, with duration: %@", asset, CFBridgingRelease(CMTimeCopyDescription(kCFAllocatorDefault, asset.duration)));
    
    AVVideoComposition *videoComposition = [self getVideoComposition:asset];
    
    NSError *error = nil;
    AVAssetReader *assetReader = [AVAssetReader assetReaderWithAsset:asset error:&error];
    

    NSMutableDictionary *outputSettings = [NSMutableDictionary dictionary];
    [outputSettings setObject:@(kCVPixelFormatType_32BGRA) forKey:(id)kCVPixelBufferPixelFormatTypeKey];
    [outputSettings setObject:[NSDictionary dictionary] forKey:(id)kCVPixelBufferIOSurfacePropertiesKey];
    [outputSettings setObject:@(1280) forKey:(id)kCVPixelBufferWidthKey];
    [outputSettings setObject:@(720) forKey:(id)kCVPixelBufferHeightKey];
    
    AVAssetReaderVideoCompositionOutput *readerVideoOutput = [AVAssetReaderVideoCompositionOutput assetReaderVideoCompositionOutputWithVideoTracks:[asset tracksWithMediaType:AVMediaTypeVideo] videoSettings:outputSettings];
    
#if ! TARGET_IPHONE_SIMULATOR
    if( [videoComposition isKindOfClass:[AVMutableVideoComposition class]] )
        [(AVMutableVideoComposition*)videoComposition setRenderScale:1.0];
#endif
    readerVideoOutput.videoComposition = videoComposition;
    readerVideoOutput.alwaysCopiesSampleData = NO;
    [assetReader addOutput:readerVideoOutput];
    
    double sampleRate = [[AVAudioSession sharedInstance] sampleRate];
    if (0 == sampleRate ) {
        // 预防获取 sampleRate 错误
        sampleRate = 44100;
    }
    int outputChannel = 1;
    
    NSArray *audioTracks = [asset tracksWithMediaType:AVMediaTypeAudio];
    AVAssetReaderAudioMixOutput *readerAudioOutput = nil;
    NSDictionary *audioSettings = [NSDictionary dictionaryWithObjectsAndKeys:
                                   [NSNumber numberWithUnsignedInt:kAudioFormatLinearPCM], AVFormatIDKey,
                                   [NSNumber numberWithInteger:outputChannel], AVNumberOfChannelsKey,
                                   [NSNumber numberWithFloat:sampleRate], AVSampleRateKey,
                                   nil];
    
    readerAudioOutput = [AVAssetReaderAudioMixOutput assetReaderAudioMixOutputWithAudioTracks:audioTracks audioSettings:audioSettings];
    readerAudioOutput.audioMix = nil;
    readerAudioOutput.alwaysCopiesSampleData = NO;
    [assetReader addOutput:readerAudioOutput];
    return assetReader;
}


- (AVMutableVideoComposition *)getVideoComposition:(AVAsset *)asset {
    AVMutableVideoComposition *videoComposition;
    AVMutableVideoCompositionInstruction * instruction;
    
    AVMutableVideoCompositionLayerInstruction *layerInstruction = nil;
    AVAssetTrack *assetVideoTrack = nil;
    
    if ([[asset tracksWithMediaType:AVMediaTypeVideo] count] != 0) {
        assetVideoTrack = [[asset tracksWithMediaType:AVMediaTypeVideo] objectAtIndex:0];
    }
    
    CGSize videoSize = CGSizeMake(1280, 720);
    CGAffineTransform transform = [self transformWithVideoAssetTrack:assetVideoTrack
                                                         rotateAngle:0
                                                        displayFrame:CGRectMake(0, 0, videoSize.width, videoSize.height)];
    
    instruction = [AVMutableVideoCompositionInstruction videoCompositionInstruction];
    instruction.timeRange = CMTimeRangeMake(kCMTimeZero, asset.duration);
    
    layerInstruction = [AVMutableVideoCompositionLayerInstruction videoCompositionLayerInstructionWithAssetTrack:assetVideoTrack];
    
    CGSize renderSize = videoSize;
    if (CGSizeEqualToSize(renderSize, CGSizeZero)) {
        renderSize = CGSizeApplyAffineTransform(assetVideoTrack.naturalSize, transform);
        renderSize = CGSizeMake(fabs(renderSize.width), fabs(renderSize.height));
    }
    
    videoComposition = [AVMutableVideoComposition videoComposition];
    videoComposition.renderSize = renderSize;
    
    float frameRate = 30;
    videoComposition.frameDuration = CMTimeMake(1000, 1000 * frameRate);
    
    NSLog(@"frameDuraton {value: %lld, timescale: %d}", videoComposition.frameDuration.value, videoComposition.frameDuration.timescale);
    
    instruction.layerInstructions = [NSArray arrayWithObject:layerInstruction];
    videoComposition.instructions = [NSArray arrayWithObject:instruction];
    
    return videoComposition;
}

- (CGAffineTransform)transformWithVideoAssetTrack:(AVAssetTrack *)videoAssetTrack
                                      rotateAngle:(CGFloat)rotateAngle
                                     displayFrame:(CGRect)displayFrame {
    
    CGFloat originAngle = atan2f(videoAssetTrack.preferredTransform.b, videoAssetTrack.preferredTransform.a);
    CGFloat totalAngle = originAngle + rotateAngle;
    
    // 将角度限定在 -M_PI ~ M_PI 内，方便下面做判断
    while (totalAngle > M_PI) {
        totalAngle -= 2 * M_PI;
    }
    while (totalAngle < -M_PI) {
        totalAngle += 2 * M_PI;
    }
    
    // rotate and translation
    CGSize naturalSize = videoAssetTrack.naturalSize;
    CGAffineTransform rotateTransform = CGAffineTransformMakeRotation(totalAngle);
    CGAffineTransform translationTransform = CGAffineTransformIdentity;
    
    if (0 <= totalAngle && totalAngle <= M_PI_2) {
        
        translationTransform = CGAffineTransformMakeTranslation(naturalSize.height * sinf(totalAngle), 0.0);
        
    } else if (M_PI_2 < totalAngle && totalAngle <= M_PI) {
        
        CGFloat angle2 = totalAngle - M_PI_2;
        
        translationTransform = CGAffineTransformMakeTranslation(naturalSize.width * sinf(angle2) + naturalSize.height * cosf(angle2), naturalSize.height * sinf(angle2));
        
    } else if (-M_PI <= totalAngle && totalAngle  <= -M_PI_2) {
        CGFloat angle2 = fabs(totalAngle) - M_PI_2;
        
        translationTransform = CGAffineTransformMakeTranslation(naturalSize.width * sinf(angle2), naturalSize.height * sinf(angle2) + naturalSize.width * cosf(angle2));
        
    } else {
        
        CGFloat angle2 = fabs(totalAngle);
        
        translationTransform = CGAffineTransformMakeTranslation(0, naturalSize.width * sinf(angle2));
    }
    
    CGAffineTransform transform = CGAffineTransformConcat(rotateTransform, translationTransform);
    
    if (CGRectEqualToRect(CGRectZero, displayFrame)) {
        CGSize renderSize = CGSizeApplyAffineTransform(naturalSize, transform);
        renderSize  = CGSizeMake(fabs(renderSize.width), fabs(renderSize.height));
        displayFrame = CGRectMake(0, 0, renderSize.width, renderSize.height);
    }
    
    // scale
    CGSize originDisplaySize = CGSizeApplyAffineTransform(videoAssetTrack.naturalSize, transform);
    originDisplaySize = CGSizeMake(fabs(originDisplaySize.width), fabs(originDisplaySize.height));
    CGFloat scaleX = 1;
    CGFloat scaleY = 1;
    scaleX = displayFrame.size.width / originDisplaySize.width;
    scaleY = displayFrame.size.height / originDisplaySize.height;
    
    CGAffineTransform scaleTransform = CGAffineTransformMakeScale(scaleX, scaleY);
    transform = CGAffineTransformConcat(transform, scaleTransform);
    
    // transform
    CGSize scaleDisplaySize = CGSizeApplyAffineTransform(videoAssetTrack.naturalSize, transform);
    scaleDisplaySize = CGSizeMake(fabs(scaleDisplaySize.width), fabs(scaleDisplaySize.height));
    CGPoint scaleCenter = CGPointMake(scaleDisplaySize.width/2, scaleDisplaySize.height/2);
    CGPoint displayCenter = CGPointMake(displayFrame.origin.x + displayFrame.size.width/2, displayFrame.origin.y + displayFrame.size.height/2);
    transform = CGAffineTransformConcat(transform, CGAffineTransformMakeTranslation(displayCenter.x - scaleCenter.x, displayCenter.y - scaleCenter.y));
    
    return transform;
}


- (void)exportMovieToPhotosAlbum:(NSArray<NSURL *> *__nonnull)urls {
    for (NSURL *url in urls) {
        NSString *pMoviePath = [url absoluteString];
        
        NSMutableString *moviePath = [NSMutableString stringWithString:pMoviePath];
        [moviePath deleteCharactersInRange:NSMakeRange(0, 7)]; // 删掉字符 file://
        
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
            if (UIVideoAtPathIsCompatibleWithSavedPhotosAlbum(moviePath)) {
                // 保存到相册
                UISaveVideoAtPathToSavedPhotosAlbum(moviePath, self, @selector(movie:didFinishSavingWithError:contextInfo:), nil);
            }
        });
    }
}

- (void)movie:(NSString *)moviePath didFinishSavingWithError:(NSError *)error contextInfo:(void *)contextInfo{
    NSLog(@"export movie to PhotosAlbum");
}

- (UIImage *)imageFromSampleBuffer:(CMSampleBufferRef) sampleBuffer
{
    // Get a CMSampleBuffer's Core Video image buffer for the media data
    CVImageBufferRef imageBuffer = CMSampleBufferGetImageBuffer(sampleBuffer);
    
    // Lock the base address of the pixel buffer
    CVPixelBufferLockBaseAddress(imageBuffer, 0);

    // Get the number of bytes per row for the pixel buffer
    void *baseAddress = CVPixelBufferGetBaseAddress(imageBuffer);

    // Get the number of bytes per row for the pixel buffer
    size_t bytesPerRow = CVPixelBufferGetBytesPerRow(imageBuffer);
    // Get the pixel buffer width and height
    size_t width = CVPixelBufferGetWidth(imageBuffer);
    size_t height = CVPixelBufferGetHeight(imageBuffer);

    // Create a device-dependent RGB color space
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();

    // Create a bitmap graphics context with the sample buffer data
    CGContextRef context = CGBitmapContextCreate(baseAddress, width, height, 8,
                                                 bytesPerRow, colorSpace, kCGBitmapByteOrder32Little | kCGImageAlphaPremultipliedFirst);
    // Create a Quartz image from the pixel data in the bitmap graphics context
    CGImageRef quartzImage = CGBitmapContextCreateImage(context);
    // Unlock the pixel buffer
    CVPixelBufferUnlockBaseAddress(imageBuffer,0);

    // Free up the context and color space
    CGContextRelease(context);
    CGColorSpaceRelease(colorSpace);

    // Create an image object from the Quartz image
    UIImage *image = [UIImage imageWithCGImage:quartzImage];

    // Release the Quartz image
    CGImageRelease(quartzImage);
    
    return (image);
}

- (UIImage *)imageAddFilter:(UIImage*)image {
    @autoreleasepool {
        CIFilter *filter = [CIFilter filterWithName:@"CISepiaTone"];
        [filter setValue:@1 forKey:kCIInputIntensityKey];
        CIImage *img = [CIImage imageWithCGImage:image.CGImage];
        [filter setValue:img forKey:kCIInputImageKey];
        CIImage* ciimage = [filter outputImage];
        CGImageRef ciimg = [context createCGImage:ciimage fromRect:ciimage.extent];
        ciimage = nil;
        UIImage *i = [UIImage imageWithCGImage:ciimg];
        CGImageRelease(ciimg);
        ciimg = NULL;
        return i;
    }
}

@end
